(ns com.aurellem.gb.rlm-assembly
  "Version of main bootstrap program that is valid output for the
   item-writer program."
  (:use (com.aurellem.gb gb-driver assembly util vbm constants))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn pc-item-writer-program
  []
  (let [limit 201
        [target-high target-low] (disect-bytes-2 pokemon-list-start)]
    (flatten
     [[0x00  ;; (item-hack) set increment stack pointer no-op
       0x1E  ;; load limit into E
       limit
       0x3F  ;; (item-hack) set carry flag no-op

       ;; load 2 into C.
       0x0E   ;; C == 1 means input-first nybble
       0x04   ;; C == 0 means input-second nybble

       0x21 ;; load target into HL
       target-low
       target-high
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0x37 ;; (item-hack) set carry flag no-op
       
       0x00 ;; (item-hack) no-op
       0xF3 ;; disable interrupts
       ;; Input Section

       0x3E ;; load 0x20 into A, to measure buttons
       0x10 

       0x00 ;; (item-hack) no-op
       0xE0 ;; load A into [FF00]
       0x00

       0xF0 ;; load 0xFF00 into A to get
       0x00 ;; button presses
       
       0xE6
       0x0F ;; select bottom four bits of A
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB8 ;; see if input is different (CP A B)

       0x00 ;; (item-hack) (INC SP)
       0x28 ;; repeat above steps if input is not different
       ;; (jump relative backwards if B != A)
       0xED ;; (literal -19) (item-hack) -19 == egg bomb (TM37)

       0x47 ;; load A into B
       
       0x0D ;; dec C
       0x37 ;; (item-hack) set-carry flag
       ;; branch based on C:
       0x20 ;; JR NZ
       23 ;; skip "input second nybble" and "jump to target" below
       
       ;; input second nybble

       0x0C ;; inc C
       0x0C ;; inc C

       0x00 ;; (item-hack) no-op
       0xE6 ;; select bottom bits
       0x0F
       0x37 ;; (item-hack) set-carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB2 ;; (OR A D) -> A

       0x22 ;; (do (A -> (HL)) (INC HL))

       0x1D ;; (DEC E)

       0x00 ;; (item-hack) 
       0x20 ;; jump back to input section if not done
       0xDA ;; literal -36 == TM 18 (counter)
       0x01 ;; (item-hack) set BC to literal (no-op)

       ;; jump to target
       0x00  ;; (item-hack) these two bytes can be anything.
       0x01 

       0x00   ;; (item-hack) no-op
       0xBF   ;; (CP A A) ensures Z
       
       0xCA   ;; (item-hack) jump if Z
       target-low
       target-high
       0x01   ;; (item-hack) will never be reached.
       
       ;; input first nybble
       0x00
       0xCB
       0x37  ;; swap nybbles on A

       0x57  ;; A -> D

       0x37  ;; (item-hack) set carry flag no-op
       0x18  ;; relative jump backwards
       0xCD  ;; literal -51 == TM05; go back to input section
       0x01  ;; (item-hack) will never reach this instruction

       ]
      (repeat 8 [0x00 0x01]);; these can be anything

      [;; jump to actual program
       0x00
       0x37  ;; (item-hack) set carry flag no-op

       0x2E  ;; 0x3A -> L
       0x3A


       0x00  ;; (item-hack) no-op
       0x26  ;; 0xD5 -> L
       0xD5  
       0x01  ;; (item-hack) set-carry BC

       0x00  ;; (item-hack) these can be anything
       0x01  

       0x00
       0xE9 ;; jump to (HL)
       ]])))



;; Specs for Main Bootstrap Program

;; Number-Input
;; Number input works using all eight buttons to
;;   spell out an 8 bit number.  The order of buttons is
;;   [:d :u :l :r :start :select :b :a]  -->  11111111
;;   [      :l    :start            :a]  -->  00101001

;;; MODES
;; There are five modes in total:
;;   MODE-SELECT
;;   SET-H
;;   SET-L
;;   WRITE
;;   JUMP

;;; MODE-SELECT
;;   The bootstrap program starts in MODE-SELECT mode.
;;   MODE-SELECT transitions to one of three modes depending
;;   on which buttons are pressed:
;;       0   :  MODE-SELECT
;;    0x67   :  SET-H
;;    0x6F   :  SET-L
;;    0x47   :  WRITE
;;    0xE9   :  JUMP

;;; SET-H
;;   SET-H sets the high 8 bits of the target address to which
;;   data will be written / the program will jump.  It expects
;;   the following:
;;
;;    Byte 0 : New Value of H
;;    Byte 1 : 0x00

;;; SET-L
;;   This mode sets the low 8 bits of the target address and has
;;   the same semantics as SET-H.

;;; WRITE-BYTES
;;   WRITE-BYTES mode writes sequences of arbitray values to
;;   arbitray memory locations. It expects you to enter a
;;   header of one byte describing how many bytes to write.

;;    Byte 0  : Number of Bytes to Write

;;   Then, you enter the number of bytes specified in Byte 0
;;   and they are written to the start address in sequence.
;;   After the last byte is written control returns to
;;   MODE-SELECT mode. The Target address will be incremented by
;;   Number of Bytes to Write once you are done writing.

;;   Example: to write the sequence [1 2 3 4] starting at
;;   the target address enter:
;;    Byte 0  : 4 (will write four bytes)
;;    Byte 3  : 1 (write 1 to 0xC01F)
;;    Byte 4  : 2 (write 2 to 0xC020)
;;    Byte 5  : 3 (write 3 to 0xC021)
;;    Byte 6  : 4 (write 4 to 0xC022)

;;; JUMP 
;;   JUMP mode jumps program control to the target address.

;;; EXAMPLE
;;   To write the infinite loop program [0x18 0xFE] to address
;;   0xC00F and then jump to said program, enter the following
;;   starting from MODE-SELECT mode.

;;    Byte 0  : 0x67 [:a :b :l :u :select]        ;; SET-H mode
;;    Byte 1  : 0xC0 [:d :u]                      ;; 0xC0 -> H
;;    Byte 2  : 0x00 []                           ;; trailer

;;    Byte 3  : 0x6F [:a :start :b :l :u :select] ;; SET-L mode
;;    Byte 4  : 0x0F [:a :start :b :select]       ;; 0x0F -> L
;;    Byte 5  : 0x00 []                           ;; trailer

;;    Byte 6  : 0x47 [:a :b :u :select]           ;; WRITE-MODE
;;    Byte 7  : 0x02 [:b]                         ;; write 2 bytes
;;    Byte 8  : 0x18 [:r :start]                  ;; assembly
;;    Byte 9  : 0xFE [:r :start :b :d :l :u :select] ;; assembly

;;    target address is now 0xC011 since we wrote 2 bytes.
;;    set it back to 0xC00F.

;;    Byte 10 : 0x6F [:a :start :b :l :u :select] ;; SET-L mode
;;    Byte 12 : 0x0F [:a :start :b :select]       ;; 0x0F -> L
;;    Byte 13 : 0x00 []                           ;; trailer

;;    Byte 14 : 0xE9                              ;; JUMP-MODE

(defn ->signed-8-bit [n]
  (if (< n 0)
    (+ 256 n) n))

(defn frame-metronome []
  (let [init [0xC5] ;; save value of BC
        timing-loop
        [0x01 ; \
         0x43 ;  |
         0xFE ;  |  load 0xFF44 into BC without repeats
         0x0C ;  |
         0x04 ; /
         0x0A] ;; (BC) -> A, now A = LY (vertical line coord)
        continue-if-144
        [0xFE
         144     ;; compare LY (in A) with 144
         0x20    ;; jump back to beginning if LY != 144 (not-v-blank)
         (->signed-8-bit
          (+ -4 (- (count timing-loop))))]
        spin-loop
        [0x05 ;; dec B, which is 0xFF
         0x20 ;; spin until B==0
         0xFD]]
    (concat init timing-loop continue-if-144 spin-loop)))

(defn frame-metronome* []
  [0x3E   ;; smallest version, but uses repeated nybbles
   0x01
   0xE0
   0xFF])

(defn frame-metronome** []
  [0x06 ;; load 0xFE into B
   0xFE
   0x04 ;; inc B, now B == FF

   0x3E ;;  RLM-debug
   0x01 ;; 1->A
  
   0x48 ;; B->C
   0x02]) ;; A->(BC) set exclusive v-blank interrupt

(defn test-frame-metronome
  "Ensure that frame-metronome ticks exactly once every frame."
  ([] (test-frame-metronome 151))
  ([steps]
     (let [inc-E [0x1C 0x18 
                  (->signed-8-bit
                   (+ -3
                      (-(count (frame-metronome)))))]
                   
           program (concat (frame-metronome) inc-E)
           count-frames
           (-> (tick (mid-game))
               (IE! 0)
               (DE! 0)
               (set-memory-range pokemon-list-start program)
               (PC! pokemon-list-start))
           E-after-moves
           (E (run-moves count-frames (repeat steps [])))]
       ;;(println "E:" E-after-moves) 
       (assert (= steps E-after-moves))
       (println "frame-count test passed.")
       count-frames)))

(defn read-user-input []
  [0x3E
   0x20 ; prepare to measure d-pad

   0x3F ; clear carry flag no-op to prevent repeated nybbles
   
   0x01 ;\
   0x01 ; |
   0xFE ; |  load 0xFF00 into BC without repeats
   0x04 ; |
   0x0D ;/
   
   0x02 
   0x0A ;; get D-pad info
   
   0xF5 ;; push AF
      
   0x3E
   0x10 ; prepare to measure buttons

   0x3F ;; clear carry flag no-op to prevent repeated nybbbles
   
   0x02
   0x0A ;; get button info
   
   0xE6 ;; select bottom bits of A
   0x0F

   0x47 ;; A->B

   0xF1 ;; pop AF

   0xE6
   0x0F ;; select bottom bits of A

   0xCB
   0x37 ;; swap A nybbles

   0xB0 ;; (or A B) -> A
 
   0x2F ;; (NOT A) -> A
   ])

(defn test-read-user-input []
  (let [program
        (concat
         (frame-metronome) (read-user-input)
         [0x5F ;; A-> E
          0x76
          0x18
          (->signed-8-bit
           (+ (- (count (read-user-input)))
              (- 4)))])
        read-input
        (-> (tick (mid-game))
            (IE! 0)
            (set-memory-range pokemon-list-start program)
            (PC! pokemon-list-start))]
    (dorun
     (for [i (range 0x100)]
       (assert (= (E (step read-input (buttons i))) i))))
     (println "tested all inputs.")
    read-input))

(def symbol-index
  (fn [symbol sequence]
    (count (take-while
            (partial not= symbol)
            sequence))))

(defn bootstrap-state-machine
  ([start-address]
     ;; Register Use:
     
     ;; ED non-volitale scratch
     
     ;; A  user-input (A MUST contain user-input for this to work!) 
     ;; HL target-address
     ;; B  bytes-to-write
     ;; C  non-volatile scratch

     ;; Modes (with codes) are:

     ;; single-action-modes:
     ;; SET-TARGET-HIGH     0x67 ;; A->H
     ;; SET-TARGET-LOW      0x6F ;; A->L
     ;; JUMP                0xE9 ;; jump to (HL)

     ;; multi-action-modes
     ;; WRITE               0x47 ;; A->B
     (let [
           input
           [0xC1  ;; pop BC so it's not volatile

            0x5F  ;; A->E
            0xAF  ;; test for output-mode (bytes-to-write > 0)
            0xB8  ;; (cp A B)
            0x7B  ;; E->A
            0x20       ;; skip to output section if
            :to-output ;; we're not in input mode 
            
            :to-be-executed

            ;; write mode to instruction-to-be-executed (pun)
            0xEA
            :to-be-executed-address

            ;; protection region -- do not queue this op for
            ;; execution if the last one was non-zero
            0x79 ;; C->A
            0xA7 ;; test A==0
            0x28
            0x04
            0xAF ;; put a no op (0x00) in to-be-executed
            0xEA ;; 
            :to-be-executed-address
            
            0x7B ;; E->A
            0x4F ;; A->C now C stores previous instruction
            0x18           ;; return
            :to-jump]
           
           output
           [:output-start ;; just a label
            0x3F ;; ;; prevent repeated nybbles
            0x54 ;;
            0x5D ;; HL->DE  \
            ;;          | This mess is here to do
            0x12 ;; A->(DE)  | 0x22 (LDI (HL), A) without
            ;;         /  any repeating nybbles
            0x05 ;; DEC bytes-to-write (B)

            0x23 ;; inc HL 
            ]
        
           symbols
           {:to-be-executed-address
            (reverse
             (disect-bytes-2
              (+ start-address
                 (symbol-index :to-be-executed input))))
            :to-be-executed 0x3F} ;; clear carry flag no-op

           program** (flatten
                      (replace
                       symbols
                       (concat input output)))
           
           resolve-internal-jumps
           {:output-start []
            :to-output
            (->signed-8-bit
             (dec
              (- (symbol-index :output-start program**)
                 (symbol-index :to-output program**))))}

           program*
           (flatten (replace resolve-internal-jumps program**))
           
           resolve-external-jumps
           {:to-jump
            (- (- (count program*)
                  (symbol-index :to-jump program*)) 1)}
           program
           (replace resolve-external-jumps program*)]
       program)))
     

(defn main-bootstrap-program
  ([] (main-bootstrap-program pokemon-list-start))
  ([start-address]
     (let [init [0xAF 0x4F 0x47] ;; 0->A; 0->C; 0->B
           header (concat (frame-metronome) (read-user-input))
           state-machine-start-address
           (+ start-address (count init) (count header))
           state-machine
           (bootstrap-state-machine state-machine-start-address)

           return-to-header
           (flatten 
            [0x18
             (->signed-8-bit
              (- (count init)
                 2 ;; this command length
                 3 ;; I have no idea why we need a 3 here
                   ;; need to investigate.
                 (count header)
                 (count state-machine)))])]
                 
       (concat init header state-machine return-to-header))))



(defn no-consecutive-repeats? [seq]
  (not (contains? (set(map - seq (rest seq))) 0)))

(defn byte->nybbles [byte]
  [(bit-shift-right byte 4) (bit-and byte 0x0F)])

(defn bootstrap-pattern
  "Given an assembly sequence, generate the keypresses required to
   create that sequence in memory using the pc-item-writer
   program. The assembly must not have any consecutive repeating
   nybbles."
  [assembly]
  (let [nybbles (flatten (map byte->nybbles assembly))
        moves (map (comp buttons (partial - 15)) nybbles)
        header (map buttons
                    (concat (repeat
                             50
                             (- 15 (first nybbles)))
                            [(first nybbles)]))]
    (assert (no-consecutive-repeats? nybbles))
    (concat header moves)))
    
;;;;;; TESTS ;;;;;;

(def set-H-mode 0x67)
(def set-L-mode 0x6F)
(def jump-mode  0xE9)
(def write-mode 0x47)


(defn bootstrap-base []
  (let [program (main-bootstrap-program pokemon-list-start)]
    ;; make sure program is valid output for item-writer
    (-> (tick (mid-game))
        (set-memory-range pokemon-list-start program)
        (PC! pokemon-list-start)
        (step [])
        (step []))))

(defn test-set-H []
  (letfn [(test-H [state n]
            (let [after
                  (-> state
                      (step (buttons set-H-mode))
                      (step (buttons n))
                      (step []))]
              ;;(println "desired H =" n "actual =" (H after))
              (assert (= n (H after)))
              after))]
    (let [result (reduce test-H (bootstrap-base) (range 0x100))]
      (println "set H test passed.")
      result)))

(defn test-write-bytes []
  (let [target-address 0xC00F
        [target-high target-low] (disect-bytes-2 target-address)
        assembly [0xF3 0x18 0xFE 0x12]
        get-mem-region  #(subvec (vec (memory %))
                                 target-address (+ target-address 20))
        before (bootstrap-base)
        after
        (-> before
            (step [])           ; make sure it can handle blanks
            (step [])           ; at the beginning.
            (step [])                      
            (step (buttons set-H-mode))     ; select set-H
            (step (buttons target-high))
            (step [])
            (step (buttons set-L-mode))
            (step (buttons target-low))
            (step [])
            (step (buttons write-mode))
            (step (buttons 4))  ; write 4 bytes
            (step (buttons (nth assembly 0)))
            (step (buttons (nth assembly 1)))
            (step (buttons (nth assembly 2)))
            (step (buttons (nth assembly 3))))]
    ;;(println "before :" (get-mem-region before))
    ;;(println "after  :" (get-mem-region after))
    ;;(assert (= assembly (take 4 (get-mem-region after))))
    (println "write-test-passed.")
    after))

(defn test-jump []
  (let [target-address 0xC00F
        [target-high target-low] (disect-bytes-2 target-address)
        post-jump
        (-> (test-write-bytes)
            (step (buttons set-H-mode))     ; select set-H
            (step (buttons target-high))
            (step [])
            (step (buttons set-L-mode))
            (step (buttons target-low))
            (step [])
            (step (buttons jump-mode)))           ; Select JUMP mode.
        program-counters
        (capture-program-counter 
         post-jump
         10000)]
    (assert (contains? (set program-counters) target-address))
    (println "jump test passed.")
    post-jump))

(defn test-no-repeated-nybbles []
  (bootstrap-pattern (main-bootstrap-program))
  (println "no-repeated-nybbles"))

(defn run-all-tests []
  (test-frame-metronome)
  (test-read-user-input)
  (test-set-H)
  (test-write-bytes)
  (test-jump)
  (test-no-repeated-nybbles)
  (println "\n all tests passed."))
